module Rakka.Authorization
    ( AuthDB
    , mkAuthDB
    , isValidPair
    )
    where

import qualified Codec.Binary.UTF8.String as UTF8
import           Control.Concurrent.STM
import           Control.Monad.Trans
import qualified Data.ByteString as B
import           Data.Map (Map)
import qualified Data.Map as M hiding (Map)
import           Data.Maybe
import           OpenSSL.EVP.Base64
import           OpenSSL.EVP.Digest
import           Rakka.SystemConfig
import           System.Directory
import           System.FilePath
import           System.IO


data AuthDB
    = AuthDB {
        adbFilePath    :: !FilePath
      , adbUserMap     :: !(TVar UserMap)
      , adbSyncRequest :: !(TVar Bool)
      }


type UserMap = Map String String


mkAuthDB :: FilePath -> IO AuthDB
mkAuthDB lsdir
    = do let path = lsdir </> "authDB"
         m   <- newTVarIO =<< loadUserMap path
         req <- newTVarIO False
         return AuthDB {
                      adbFilePath    = path
                    , adbUserMap     = m
                    , adbSyncRequest = req
                    }


isValidPair :: MonadIO m => AuthDB -> String -> String -> m Bool
isValidPair adb name pass
    = liftIO $ do sha1 <- return . fromJust =<< getDigestByName "SHA1"
                  let hash = digestBS sha1 $ B.pack $ UTF8.encode pass
                  atomically $ do m <- readTVar (adbUserMap adb)
                                  return (M.lookup name m == Just hash)


loadUserMap :: FilePath -> IO UserMap
loadUserMap path
    = do exist <- doesFileExist path
         m     <- if exist then
                      readFile path
                      >>=
                      return . M.fromList . map decodePair . fromJust . deserializeStringPairs
                  else
                      return M.empty
         sha1  <- return . fromJust =<< getDigestByName "SHA1"
         return (initMap sha1 m)
    where
      decodePair :: (String, String) -> (String, String)
      decodePair (name, b64Hash)
          = (UTF8.decodeString name, decodeBase64 b64Hash)

      initMap :: Digest -> UserMap -> UserMap
      initMap sha1 m
          | M.null m  = let name = "root"
                            hash = digest sha1 ""
                        in
                          M.singleton name hash
          | otherwise = m
